from __future__ import annotations

from typing import Any, Iterable
from itertools import islice
from check_link import Link, check_all

import ckan.plugins.toolkit as tk
from ckan.logic import validate
from ckan.lib.search.query import solr_literal

from ckanext.toolbelt.decorators import Collector

from .. import schema


action, get_actions = Collector("check_link").split()


@action
@validate(schema.url_check)
def url_check(context, data_dict):
    tk.check_access("check_link_url_check", context, data_dict)

    try:
        result = check_all(map(Link, data_dict["url"]))
    except ValueError:
        raise tk.ValidationError({"url": ["Must be a valid URL"]})

    reports = [
        {
            "url": link.link,
            "state": link.state.name,
            "code": link.code,
            "reason": link.reason,
            "explanation": link.details,
        }
        for link in result
    ]

    if data_dict["save"]:
        _save_reports(context, reports, data_dict["clear_available"])

    return reports


@action
@validate(schema.resource_check)
def resource_check(context, data_dict):
    tk.check_access("check_link_resource_check", context, data_dict)
    resource = tk.get_action("resource_show")(context, data_dict)

    result = tk.get_action("check_link_url_check")(
        context, {"url": [resource["url"]]}
    )

    report = dict(result[0], resource_id=resource["id"], package_id=resource["package_id"])

    if data_dict["save"]:
        _save_reports(context, [report], data_dict["clear_available"])

    return report


@action
@validate(schema.package_check)
def package_check(context, data_dict):
    tk.check_access("check_link_package_check", context, data_dict)
    return _search_check(
        context,
        "res_url:* (id:{0} OR name:{0})".format(solr_literal(data_dict["id"])),
        data_dict,
    )["reports"]


@action
@validate(schema.organization_check)
def organization_check(context, data_dict):
    tk.check_access("check_link_organization_check", context, data_dict)

    return _search_check(
        context,
        "res_url:* owner_org:{}".format(solr_literal(data_dict["id"])),
        data_dict,
    )["reports"]


@action
@validate(schema.group_check)
def group_check(context, data_dict):
    tk.check_access("check_link_group_check", context, data_dict)

    return _search_check(
        context, "res_url:* groups:{}".format(solr_literal(data_dict["id"])), data_dict
    )["reports"]


@action
@validate(schema.user_check)
def user_check(context, data_dict):
    tk.check_access("check_link_user_check", context, data_dict)

    return _search_check(
        context,
        "res_url:* creator_user_id:{}".format(solr_literal(data_dict["id"])),
        data_dict,
    )["reports"]


@action
@validate(schema.search_check)
def search_check(context, data_dict):
    tk.check_access("check_link_search_check", context, data_dict)

    return _search_check(context, data_dict["fq"], data_dict)["reports"]


def _search_check(context, fq: str, data_dict: dict[str, Any]):
    params = {
        "fq": fq,
        "start": data_dict["start"],
        "include_drafts": data_dict["include_drafts"],
        "include_deleted": data_dict["include_deleted"],
        "include_private": data_dict["include_private"],
    }

    pairs = [
        ({"resource_id": res["id"], "package_id": pkg["id"]}, res["url"])
        for pkg in islice(_iterate_search(context, params), data_dict["rows"])
        for res in pkg["resources"]
    ]

    if not pairs:
        return {"reports", []}

    patches, urls = zip(*pairs)

    result = tk.get_action("check_link_url_check")(
        context, {"url": urls}
    )

    reports = [dict(report, **patch) for patch, report in zip(patches, result)]
    if data_dict["save"]:
        _save_reports(context, reports, data_dict["clear_available"])

    return {
        "reports": reports,
    }


def _iterate_search(context, params: dict[str, Any]):
    params.setdefault("start", 0)

    while True:
        pack = tk.get_action("package_search")(context.copy(), params)
        if not pack["results"]:
            return

        yield from pack["results"]

        params["start"] += len(pack["results"])


def _save_reports(context, reports: Iterable[dict[str, Any]], clear: bool):
        save = tk.get_action("check_link_report_save")
        delete = tk.get_action("check_link_report_delete")

        for report in reports:
            if clear and report["state"] == "available":
                try:
                    delete(context.copy(), report)
                except tk.ObjectNotFound:
                    pass
            else:
                save(context.copy(), report)



from __future__ import annotations

from ckan import authz
import ckan.plugins.toolkit as tk
from ckanext.toolbelt.decorators import Collector

CONFIG_ALLOW_USER = "ckanext.check_link.user_can_check_url"
DEFAULT_ALLOW_USER = False

auth, get_auth_functions = Collector("check_link").split()


@auth
def url_check(context, data_dict):
    allow_user_checks = tk.asbool(
        tk.config.get(
            CONFIG_ALLOW_USER,
            DEFAULT_ALLOW_USER,
        )
    )
    return {"success": allow_user_checks and not authz.auth_is_anon_user(context)}


@auth
def resource_check(context, data_dict):
    return authz.is_authorized("resource_show", context, data_dict)


@auth
def package_check(context, data_dict):
    return authz.is_authorized("package_show", context, data_dict)


@auth
def organization_check(context, data_dict):
    return authz.is_authorized("organization_show", context, data_dict)


@auth
def group_check(context, data_dict):
    return authz.is_authorized("group_show", context, data_dict)


@auth
def user_check(context, data_dict):
    return authz.is_authorized("user_show", context, data_dict)


@auth
def search_check(context, data_dict):
    return authz.is_authorized("package_search", context, data_dict)


@auth
def report_save(context, data_dict):
    return authz.is_authorized("sysadmin", context, data_dict)


@auth
def report_show(context, data_dict):
    return authz.is_authorized("sysadmin", context, data_dict)


@auth
def report_search(context, data_dict):
    return authz.is_authorized("sysadmin", context, data_dict)


@auth
def report_delete(context, data_dict):
    return authz.is_authorized("sysadmin", context, data_dict)


@auth
def view_report_page(context, data_dict):
    return authz.is_authorized("sysadmin", context, data_dict)



import pytest

from ckan.tests.helpers import call_action
import ckan.plugins.toolkit as tk


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestSave:
    def test_resource_id_must_be_real(self, resource, faker):
        with pytest.raises(tk.ValidationError):
            call_action(
                "check_link_report_save",
                url=faker.url(),
                state="unknown",
                resource_id=faker.uuid4(),
            )

        call_action(
            "check_link_report_save",
            url=faker.url(),
            state="unknown",
            resource_id=resource["id"],
        )

    def test_url_is_mandatory(self, resource, faker):
        with pytest.raises(tk.ValidationError):
            call_action(
                "check_link_report_save", state="unknown", resource_id=resource["id"]
            )

    def test_state_is_mandatory(self, resource, faker):
        with pytest.raises(tk.ValidationError):
            call_action(
                "check_link_report_save", url=faker.url(), resource_id=resource["id"]
            )

    def test_update_existing_by_resource_id(self, resource, report_factory):
        report = report_factory(resource_id=resource["id"], state="unknown")

        updated = call_action(
            "check_link_report_save",
            url=report["url"],
            resource_id=resource["id"],
            state="updated",
        )

        assert updated["id"] == report["id"]
        assert updated["state"] == "updated"

    def test_update_existing_by_url(self, report_factory):
        report = report_factory(resource_id=None, state="unknown")

        updated = call_action(
            "check_link_report_save", url=report["url"], state="updated"
        )

        assert updated["id"] == report["id"]
        assert updated["state"] == "updated"

    def test_update_existing_by_id(self, report_factory, faker):
        report = report_factory(resource_id=None, state="unknown")

        updated = call_action(
            "check_link_report_save", url=faker.url(), state="updated", id=report["id"]
        )

        assert updated["id"] == report["id"]
        assert updated["state"] == "updated"


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestShow:
    def test_shown_by_id(self, report):
        assert (
            call_action("check_link_report_show", id=report["id"])["id"] == report["id"]
        )
        assert (
            call_action("check_link_report_show", id=report["id"])["id"] == report["id"]
        )

    def test_shown_by_resource_id(self, report_factory):
        with_resource = report_factory()
        assert (
            call_action(
                "check_link_report_show", resource_id=with_resource["resource_id"]
            )["id"]
            == with_resource["id"]
        )

    def test_shown_by_url(self, report_factory):
        with_resource = report_factory()
        without_resource = report_factory(resource_id=None)
        assert (
            call_action("check_link_report_show", url=without_resource["url"])["id"]
            == without_resource["id"]
        )

        with pytest.raises(tk.ObjectNotFound):
            assert call_action("check_link_report_show", url=with_resource["url"])


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestDelete:
    def test_delete(self, report):
        call_action("check_link_report_delete", id=report["id"])
        with pytest.raises(tk.ObjectNotFound):
            assert call_action("check_link_report_show", id=report["id"])


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestSearch:
    def test_limit(self, report_factory):
        report_factory.create_batch(10)

        result = call_action("check_link_report_search", limit=5)
        assert result["count"] == 10
        assert len(result["results"]) == 5

        result = call_action("check_link_report_search", limit=5, offset=8)
        assert result["count"] == 10
        assert len(result["results"]) == 2
